import {Fuel} from "$lib/fuel";
import {Grid} from "$lib/grid";
import type {ObjectValues} from "$lib/components/object-values";

export enum ThrusterSize {
    Small = "smallThruster",
    Large = "largeThruster",
    Huge = "hugeThruster",
}

export type ThrustDensity = {
    thrust: number;
    density: number;
}

export class ThrustCurve {
    private min: ThrustDensity;
    private max: ThrustDensity;

    constructor(min: ThrustDensity, max: ThrustDensity) {
        this.min = min;
        this.max = max;
    }

    getEffectiveness(airDensity: number): number {
        if (airDensity >= this.max.density) {
            return this.max.thrust;
        }
        else if (airDensity <= this.min.density) {
            return this.min.thrust;
        }
        else {
            return (1 - this.getPointOnCurve(airDensity)) * this.min.thrust + this.getPointOnCurve(airDensity) * this.max.thrust
        }
    }

    private getPointOnCurve(airDensity: number): number {
        return (airDensity-this.min.density)/(this.max.density - this.min.density);
    }
}

type ThrusterTypeDetails = {
    key: string,
    name: string,
    curve: ThrustCurve,
    fuel: Fuel,
};

export const THRUSTER_TYPE_LIST: { [key: string]: ThrusterTypeDetails } = {
    Atmospheric: {
        key: "Atmospheric",
        name: "atmosphericThruster",
        curve: new ThrustCurve({density: 0.3, thrust: 0}, {density: 1, thrust: 1}),
        fuel: Fuel.Electricity,
    },
    Ion: {
        key: "Ion",
        name: "ionThruster",
        curve: new ThrustCurve({density: 0, thrust: 1}, {density: 1, thrust: 0.3}),
        fuel: Fuel.Electricity,
    },
    Hydrogen: {
        key: "Hydrogen",
        name: "hydrogenThruster",
        curve: new ThrustCurve({density: 0, thrust: 1}, {density: 1, thrust: 1}),
        fuel: Fuel.Hydrogen,
    },
    Fusion: {
        key: "Fusion",
        name: "fusionThruster",
        curve: new ThrustCurve({density: 0, thrust: 1}, {density: 1, thrust: 0.5}),
        fuel: Fuel.Deuterium,
    },
}

export class ThrusterType {
    details: ThrusterTypeDetails = $state(THRUSTER_TYPE_LIST.Atmospheric);

    constructor(type: ThrusterTypeDetails) {
        if (typeof type === "undefined") {
            throw("ThrusterType undefined, cannot create Object");
        }
        this.details = type;
    }

    equals(obj: ThrusterType) {
        let equals: Array<boolean> = [];
        equals[0] = this.details.key === obj.details.key;
        equals[1] = this.details.name === obj.details.name;
        equals[2] = this.details.curve === obj.details.curve;
        equals[3] = this.details.fuel === obj.details.fuel;
        return !equals.includes(false);
    }
}

type ThrusterDetails = {
    key: string,
    name: string,
    type: ThrusterType,
    size: ThrusterSize,
    fuelConsumption: Map<Grid, number>;
    thrust: Map<Grid, number>;
}

export const THRUSTER_LIST: {[key: string]: ThrusterDetails } = {
    SmallAtmos: {
        key: "SmallAtmos",
        name: "smallAtmos",
        type: new ThrusterType(THRUSTER_TYPE_LIST.Atmospheric),
        size: ThrusterSize.Small,
        fuelConsumption: new Map([
            [Grid.Small, 0],
            [Grid.Large, 0],
        ]),
        thrust: new Map([
            [Grid.Small, 65000],
            [Grid.Large, 350000],
        ])
    },
    LargeAtmos: {
        key: "LargeAtmos",
        name: "largeAtmos",
        type: new ThrusterType(THRUSTER_TYPE_LIST.Atmospheric),
        size: ThrusterSize.Large,
        fuelConsumption: new Map([
            [Grid.Small, 0],
            [Grid.Large, 0],
        ]),
        thrust: new Map([
            [Grid.Small, 340000],
            [Grid.Large, 4500000],
        ])
    },

    SmallIon: {
        key: "SmallIon",
        name: "smallIon",
        type: new ThrusterType(THRUSTER_TYPE_LIST.Ion),
        size: ThrusterSize.Small,
        fuelConsumption: new Map([
            [Grid.Small, 0],
            [Grid.Large, 0],
        ]),
        thrust: new Map([
            [Grid.Small, 14400],
            [Grid.Large, 345600],
        ])
    },
    LargeIon: {
        key: "LargeIon",
        name: "largeIon",
        type: new ThrusterType(THRUSTER_TYPE_LIST.Ion),
        size: ThrusterSize.Large,
        fuelConsumption: new Map([
            [Grid.Small, 0],
            [Grid.Large, 0],
        ]),
        thrust: new Map([
            [Grid.Small, 172800],
            [Grid.Large, 4320000],
        ])
    },

    SmallHydrogen: {
        key: "SmallHydrogen",
        name: "smallHydrogen",
        type: new ThrusterType(THRUSTER_TYPE_LIST.Hydrogen),
        size: ThrusterSize.Small,
        fuelConsumption: new Map([
            [Grid.Small, 0],
            [Grid.Large, 0],
        ]),
        thrust: new Map([
            [Grid.Small, 98400],
            [Grid.Large, 1080000],
        ])
    },
    LargeHydrogen: {
        key: "LargeHydrogen",
        name: "largeHydrogen",
        type: new ThrusterType(THRUSTER_TYPE_LIST.Hydrogen),
        size: ThrusterSize.Large,
        fuelConsumption: new Map([
            [Grid.Small, 0],
            [Grid.Large, 0],
        ]),
        thrust: new Map([
            [Grid.Small, 480000],
            [Grid.Large, 7200000],
        ])
    },

    SmallFusion: {
        key: "SmallFusion",
        name: "smallFusion",
        type: new ThrusterType(THRUSTER_TYPE_LIST.Fusion),
        size: ThrusterSize.Small,
        fuelConsumption: new Map([
            [Grid.Small, 0],
            [Grid.Large, 0],
        ]),
        thrust: new Map([
            [Grid.Small, 210000],
            [Grid.Large, 9040000],
        ])
    },
    LargeFusion: {
        key: "LargeFusion",
        name: "largeFusion",
        type: new ThrusterType(THRUSTER_TYPE_LIST.Fusion),
        size: ThrusterSize.Large,
        fuelConsumption: new Map([
            [Grid.Small, 0],
            [Grid.Large, 0],
        ]),
        thrust: new Map([
            [Grid.Small, 1030000],
            [Grid.Large, 33780000],
        ])
    },
    HugeFusion: {
        key: "HugeFusion",
        name: "hugeFusion",
        type: new ThrusterType(THRUSTER_TYPE_LIST.Fusion),
        size: ThrusterSize.Huge,
        fuelConsumption: new Map([
            [Grid.Small, 0],
        ]),
        thrust: new Map([
            [Grid.Small, 9040000],
        ])
    }
}

export class Thruster {
    details: ThrusterDetails;

    constructor(thruster: ThrusterDetails) {
        if (typeof thruster === "undefined") {
            throw("Thruster undefined, cannot create Object");
        }
        this.details = thruster;
    }

    equals(obj: Thruster): boolean {
        let equals: Array<boolean> = [];
        equals[0] = this.details.key === obj.details.key;
        equals[1] = this.details.name === obj.details.name;
        equals[2] = this.details.type === obj.details.type;
        equals[3] = this.details.size === obj.details.size;
        equals[4] = this.details.fuelConsumption === obj.details.fuelConsumption;
        equals[5] = this.details.thrust === obj.details.thrust;
        return !equals.includes(false);
    }

    getThrust(grid: Grid, atmosphereDensity: number): number {
        let efficiencyCoefficient: number = this.details.type.details.curve.getEffectiveness(atmosphereDensity);
        let thruster: ThrusterDetails | undefined = THRUSTER_LIST[this.details.key];
        if (thruster === undefined) {
            throw(`Thruster ${this.details.key} undefined, cannot retrieve thrust values`);
        }
        let thrust: number | undefined = thruster.thrust.get(grid);
        if (thrust === undefined) {
            console.warn(`Thruster ${this.details.key} has no thrust values. Defaulting to 0`);
            thrust = 0;
        }
        return thrust * efficiencyCoefficient;
    }

    getMaxThrust(grid: Grid): number {
        let thruster: ThrusterDetails | undefined = THRUSTER_LIST[this.details.key];
        if (thruster === undefined) {
            throw("Thruster undefined, cannot retrieve thrust values");
        }
        let thrust: number | undefined = thruster.thrust.get(grid);
        if (thrust === undefined) {
            console.warn(`Thruster ${this.details.key} has no thrust values. Defaulting to 0`);
            thrust = 0;
        }
        return thrust;
    }
    
    static getByType(type: ThrusterType, grid: Array<Grid> = Object.values(Grid)): Array<ThrusterDetails> {
        let thrusters: Array<ThrusterDetails> = [];
        Object.values(THRUSTER_LIST).forEach((thruster) => {
            if (thruster.type.details.key === type.details.key) {
                thrusters.push(thruster);
            }
        })
        return thrusters;
    }
}